@@ -26,6 +26,7 @@
 #include <apr_file_io.h>
 #include <apr_hash.h>
 #include <apr_hooks.h>
+#include <apr_poll.h>
 #include <apr_pools.h>
 #include <apr_tables.h>
 
@@ -60,6 +61,8 @@ void redwax_set_module_config(redwax_conf_vector_t *cv, const module *m,
 #define redwax_set_module_config(v,m,val)   \
     ((((void **)(v))[(m)->module_index]) = (val))
 
+typedef struct redwax_metadata_t redwax_metadata_t;
+
 typedef struct redwax_filter_t {
     int filter_applied;
 } redwax_filter_t;
@@ -97,6 +100,11 @@ typedef enum redwax_expiry_e {
     REDWAX_EXPIRY_IGNORE_CHAIN
 } redwax_expiry_e;
 
+typedef enum redwax_dane_e {
+    REDWAX_DANE_CHECK = 0,
+    REDWAX_DANE_IGNORE
+} redwax_dane_e;
+
 typedef struct redwax_tool_t {
     apr_pool_t *pool;
     apr_pool_t *tpool;
@@ -118,6 +126,8 @@ typedef struct redwax_tool_t {
     apr_array_header_t *crls_out;
     apr_array_header_t *keys_in;
     apr_array_header_t *keys_out;
+    apr_array_header_t *dns_requests;
+    apr_array_header_t *tlsa_qnames;
     apr_hash_t *emails;
     apr_hash_t *hostnames;
     apr_hash_t *ips;
@@ -130,6 +140,7 @@ typedef struct redwax_tool_t {
     apr_hash_t *cert_relationships;
     const char *verify_param;
     const char *verify_date;
+    const char *purpose;
     const char *secret_suffix_in;
     const char *secret_suffix_out;
     const char *secret_token_in;
@@ -140,6 +151,8 @@ typedef struct redwax_tool_t {
     const char *group_in;
     const char *group_out;
     const char *calendar_alarm;
+    const char *dane_basename;
+    apr_pollcb_t *poll;
     redwax_filter_t filter;
     redwax_nss_t nss_out;
     redwax_pkcs11_t pkcs11_in;
@@ -148,6 +161,13 @@ typedef struct redwax_tool_t {
     redwax_format_e format;
     redwax_order_e order;
     redwax_expiry_e expiry;
+    redwax_dane_e dane;
+    int dane_bogus;
+    int dane_insecure;
+    int dane_nxdomain;
+    int dane_malformed;
+    int dane_record;
+    int poll_work;
     int current;
     int cert_out;
     int chain_out;
@@ -247,6 +267,7 @@ typedef struct redwax_certificate_t {
 
 typedef struct redwax_crl_t {
     apr_pool_t *pool;
+    redwax_conf_vector_t *per_module;
     const char *header;
     const unsigned char *der;
     apr_size_t len;
@@ -331,6 +352,53 @@ typedef struct redwax_offset_t {
     apr_size_t size;
 } redwax_offset_t;
 
+typedef struct redwax_dns_t redwax_dns_t;
+
+typedef apr_status_t(* redwax_dns_cb_t) (redwax_tool_t *r, redwax_dns_t *dns);
+
+typedef struct redwax_rdata_tlsa_t {
+    const char *usage_name;
+    const char *selector_name;
+    const char *mtype_name;
+    unsigned char usage;
+    unsigned char selector;
+    unsigned char mtype;
+    unsigned char *data;
+    int len;
+} redwax_rdata_tlsa_t;
+
+typedef struct redwax_rdata_t {
+    unsigned char *data;
+    int len;
+    union {
+        redwax_rdata_tlsa_t tlsa;
+    } rr;
+} redwax_rdata_t;
+
+typedef struct redwax_dns_t {
+    redwax_tool_t *r;
+    const char *basename;
+    const char *qname;
+    int rrtype;
+    int rrclass;
+    apr_array_header_t *rdata;
+    const char *canonname;
+    int havedata;
+    int nxdomain;
+    int secure;
+    int bogus;
+    const char *why_bogus;
+    redwax_dns_cb_t cb;
+    void *ctx;
+} redwax_dns_t;
+
+typedef apr_status_t(* redwax_pollfd_cb_t) (redwax_tool_t *r, void *ctx, apr_pollfd_t *descriptor);
+
+typedef struct redwax_pollfd_t {
+    redwax_pollfd_cb_t cb;
+    void *ctx;
+} redwax_pollfd_t;
+
 typedef enum redwax_token_escape_e {
     REDWAX_TOKEN_NOESCAPE = 0,
     REDWAX_TOKEN_WASESCAPE,
@@ -397,6 +465,14 @@ apr_status_t redwax_complete_file(redwax_tool_t *r, const char *arg,
 #define MAXHOSTNAMELEN 256
 #endif
 
+#ifndef APR_WANT_READ
+#define APR_WANT_READ                  (APR_UTIL_START_STATUS + 115)
+#endif
+
+#ifndef APR_WANT_WRITE
+#define APR_WANT_WRITE                 (APR_UTIL_START_STATUS + 116)
+#endif
+
 /*
  * Comparisons.
  */
@@ -708,6 +784,14 @@ APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, set_verify_param,
 APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, set_verify_date,
         (redwax_tool_t *r, const char *arg));
 
+/**
+ * Hook to complete the verification expiry.
+ *
+ * @param r The redwax-tool context.
+ */
+APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, complete_verify_expiry,
+        (redwax_tool_t *r, apr_hash_t *purposes));
+
 /**
  * Hook to set the verification expiry.
  *
@@ -716,6 +800,70 @@ APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, set_verify_date,
 APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, set_verify_expiry,
         (redwax_tool_t *r, const char *arg));
 
+/**
+ * Hook to complete verification dane.
+ *
+ * @param r The redwax-tool context.
+ */
+APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, complete_verify_dane,
+        (redwax_tool_t *r, apr_hash_t *purposes));
+
+/**
+ * Hook to set the verification dane policy.
+ *
+ * @param r The redwax-tool context.
+ */
+APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, set_verify_dane,
+        (redwax_tool_t *r, const char *arg));
+
+/**
+ * Hook to complete purposes.
+ *
+ * @param r The redwax-tool context.
+ */
+APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, complete_purpose,
+        (redwax_tool_t *r, apr_hash_t *purposes));
+
+/**
+ * Hook to set the purposes.
+ *
+ * @param r The redwax-tool context.
+ */
+APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, set_purpose,
+        (redwax_tool_t *r, const char *arg));
+
+/**
+ * Hook to set the tlsa.
+ *
+ * @param r The redwax-tool context.
+ */
+APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, set_tlsa,
+        (redwax_tool_t *r, const char *arg));
+
+/**
+ * Hook to set the smimea.
+ *
+ * @param r The redwax-tool context.
+ */
+APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, set_smimea,
+        (redwax_tool_t *r, const char *arg));
+
+/**
+ * Hook to set the DNS server address.
+ *
+ * @param r The redwax-tool context.
+ */
+APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, set_dns_server,
+        (redwax_tool_t *r, const char *arg));
+
+/**
+ * Hook to set the DNS trust anchor.
+ *
+ * @param r The redwax-tool context.
+ */
+APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, set_dns_trust_anchor,
+        (redwax_tool_t *r, const char *arg));
+
 /**
  * Hook to search for intermediate and root certificates.
  *
@@ -768,5 +916,52 @@ APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, normalise_key,
 APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, normalise_certificate,
         (redwax_tool_t *r, redwax_certificate_t *cert, int index));
 
+/**
+ * Hook to add DNS entries to metadata.
+ *
+ * @param r The redwax-tool context.
+ * @param m The metadata tree.
+ * @param cert The certificate.
+ */
+APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, add_dns_metadata,
+        (redwax_tool_t *r, redwax_metadata_t *m, const redwax_certificate_t *cert));
+
+/**
+ * Hook to process DNS results.
+ *
+ * @param r The redwax-tool context.
+ * @param dns The dns result.
+ * @param rdata The dns record.
+ */
+APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, process_dns,
+        (redwax_tool_t *r, redwax_dns_t *dns, redwax_rdata_t *rdata));
+
+/**
+ * Hook to process TLSA results.
+ *
+ * @param r The redwax-tool context.
+ * @param dns The dns result.
+ */
+APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, process_tlsa,
+        (redwax_tool_t *r, redwax_dns_t *dns));
+
+/**
+ * Hook to process SMIMEA results.
+ *
+ * @param r The redwax-tool context.
+ * @param dns The dns result.
+ */
+APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, process_smimea,
+        (redwax_tool_t *r, redwax_dns_t *dns));
+
+/**
+ * Hook to run the pre-filter poll.
+ *
+ * This is where DNS lookups and TLS connection attempts happen.
+ *
+ * @param r The redwax-tool context.
+ */
+APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, filter_poll,
+        (redwax_tool_t *r));
 
 #endif
