@@ -42,6 +42,8 @@ typedef struct redwax_libical_vcalendar_t {
     icalcomponent *cal;
 } redwax_libical_vcalendar_t;
 
+module libical_module;
+
 static apr_status_t redwax_libical_initialise(redwax_tool_t *r)
 {
 
@@ -698,4 +700,16 @@ void redwax_add_default_libical_hooks()
     rt_hook_set_calendar_alarm(redwax_libical_set_calendar_alarm, NULL, NULL, APR_HOOK_MIDDLE);
 }
 
+#else
+
+void redwax_add_default_libical_hooks()
+{
+}
+
 #endif
+
+REDWAX_DECLARE_MODULE(libical) =
+{
+    STANDARD_MODULE_STUFF,
+    redwax_add_default_libical_hooks                   /* register hooks */
+};
